use subprocess::Exec;
use yansi::{Color, Paint};

use crate::{
    NO_COMMIT_REGEX, no_commit_amount, precommit::rust_pre_commit, read_stdout, show_rg,
    todos_amount,
};

/// Check if the current repository has a remote set up
pub fn has_remote() -> bool {
    let e = Exec::cmd("git").arg("remote").arg("-v");
    let str = read_stdout(e);
    !str.is_empty()
}

/// Create a new branch and switch to it in the current repository
pub fn create_new_branch(branch: &str) {
    let mut git = Exec::cmd("git")
        .arg("checkout")
        .arg("-b")
        .arg(branch)
        .popen()
        .unwrap();
    git.wait().unwrap();
    if has_remote() {
        let mut git = Exec::cmd("git")
            .arg("push")
            .arg("-u")
            .arg("origin")
            .arg(branch)
            .popen()
            .unwrap();
        git.wait().unwrap();
    }
}

/// Push the current repository to remote
pub fn push(force: bool) {
    let mut cmd = Exec::cmd("git")
        .arg("push")
        .arg("--set-upstream")
        .arg("origin");

    if force {
        cmd = cmd.arg("--force-with-lease");
    }

    cmd.popen().unwrap().wait().unwrap();
}

/// Delete a branch from the current repository
pub fn delete_branch(branch: &str) {
    let mut git = Exec::cmd("git")
        .arg("branch")
        .arg("-d")
        .arg(branch)
        .popen()
        .unwrap();
    git.wait().unwrap();
    if has_remote() {
        let mut git = Exec::cmd("git")
            .arg("push")
            .arg("origin")
            .arg("--delete")
            .arg(branch)
            .popen()
            .unwrap();
        git.wait().unwrap();
    }
}

/// Get the title of the last commit
pub fn last_commit() -> String {
    let e = Exec::cmd("git").arg("log").arg("-1").arg("--pretty=%B");
    read_stdout(e)
}

/// Get all files affected by last commit
pub fn last_commit_files() -> Vec<String> {
    let file_str = read_stdout(
        Exec::cmd("git")
            .arg("diff-tree")
            .arg("--no-commit-id")
            .arg("--name-only")
            .arg("-r")
            .arg("HEAD"),
    );

    file_str.split('\n').map(|x| x.to_string()).collect()
}

/// Add a file to git
pub fn git_add(file: &str) {
    Exec::cmd("git")
        .arg("add")
        .arg(file)
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}

/// Get a Vec of programming languages used in this repository
pub fn get_languages() -> Vec<String> {
    let e: Exec = Exec::cmd("onefetch").arg("-o").arg("json");
    let json_str = read_stdout(e);
    if let Result::<serde_json::Value, _>::Ok(json) = serde_json::from_str(&json_str) {
        let languages = (|| {
            let fields = json.as_object()?.get("infoFields")?.as_array()?;
            let langs = fields
                .iter()
                .find(|x| x.as_object().unwrap().contains_key("LanguagesInfo"))?;
            let langs = langs
                .as_object()?
                .get("LanguagesInfo")?
                .as_object()?
                .get("languagesWithPercentage")?
                .as_array()?;
            Some(
                langs
                    .iter()
                    .map(|x| {
                        x.as_object()
                            .unwrap()
                            .get("language")
                            .unwrap()
                            .as_str()
                            .unwrap()
                            .to_string()
                    })
                    .collect(),
            )
        })();

        languages.unwrap_or_default()
    } else {
        Vec::new()
    }
}

/// Commit changes to the repository.
pub fn commit(all: bool, done: bool, msg: &str) {
    // Work In Progress Save Commit
    if last_commit().as_str() == "WIP" {
        // Get files affected by commit
        let last_commit_files = last_commit_files();

        // Reset last commit
        Exec::cmd("git")
            .arg("reset")
            .arg("HEAD~1")
            .popen()
            .unwrap()
            .wait()
            .unwrap();

        // Readd files affected by commit
        for file in last_commit_files {
            git_add(&file);
        }
    }

    // Check TODOs
    let todos = todos_amount();

    if todos > 0 {
        println!(
            "{}: Repository still contains {todos} TODO entries.",
            "Warning".paint(Color::Yellow)
        );

        if done {
            println!(
                "{}",
                "💀 Make sure to close them before commiting or we can't rest in peace..."
                    .paint(Color::Red)
            );
            std::process::exit(1);
        }
    }

    // Impossible commits
    if no_commit_amount() > 0 {
        println!("{}", "Unable to commit because of:".paint(Color::Red));
        show_rg(NO_COMMIT_REGEX);
        std::process::exit(1);
    }

    // Laguage specific pre-commit hooks
    for lang in get_languages() {
        match lang.as_str() {
            "Rust" => {
                rust_pre_commit();
            }
            _ => {}
        }
    }

    if all {
        Exec::cmd("git")
            .arg("add")
            .arg("-A")
            .popen()
            .unwrap()
            .wait()
            .unwrap();
    }

    Exec::cmd("git")
        .arg("commit")
        .arg("-m")
        .arg(msg)
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}

/// Fetch remote state
pub fn fetch() {
    Exec::cmd("git")
        .arg("fetch")
        .arg("--prune")
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}

/// Pull remote changes
pub fn pull() {
    Exec::cmd("git")
        .arg("pull")
        .arg("--prune")
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}

/// Bootstrap a git repository using a base template
pub fn bootstrap(base: &str, name: &str) {
    Exec::cmd("git")
        .arg("clone")
        .arg(base)
        .arg(name)
        .popen()
        .unwrap()
        .wait()
        .unwrap();
    std::fs::remove_dir_all(std::path::Path::new(name).join(".git")).unwrap();
    Exec::cmd("git")
        .arg("init")
        .arg("--quiet")
        .cwd(std::path::Path::new(name))
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}
