use git::{bootstrap, commit, create_new_branch, delete_branch, fetch, git_add, pull, push};
use gitmoji::{CommitMessage, select_gitmoji};
use std::io::Read;
use subprocess::{Exec, Redirection};

mod args;
mod git;
mod gitmoji;
mod precommit;

pub fn read_stdout(e: Exec) -> String {
    let mut p = e.stdout(Redirection::Pipe).popen().unwrap();
    let mut str = String::new();
    p.stdout.as_mut().unwrap().read_to_string(&mut str).unwrap();
    str.trim().to_string()
}

const TODO_REGEX: &str = r" (todo|unimplemented|refactor|wip|fix)(:|!|\n|\ :)";
const NO_COMMIT_REGEX: &str = r"(NOCOMMIT|ENSURE: )";

pub fn no_commit_amount() -> u64 {
    rg_matches(NO_COMMIT_REGEX)
}

pub fn todos_amount() -> u64 {
    rg_matches(TODO_REGEX)
}

pub fn rg_matches(regex: &str) -> u64 {
    let mut cmd = Exec::cmd("rg")
        .arg("-i")
        .arg("--json")
        .arg("--multiline")
        .arg(regex)
        .stdout(Redirection::Pipe)
        .popen()
        .unwrap();
    cmd.wait().unwrap();
    let mut str = String::new();
    cmd.stdout
        .as_mut()
        .unwrap()
        .read_to_string(&mut str)
        .unwrap();
    let last_line = str.lines().last().unwrap();
    let val: serde_json::Value = serde_json::from_str(last_line).unwrap();
    let ret = (|| {
        val.as_object()
            .unwrap()
            .get("data")?
            .as_object()?
            .get("stats")?
            .as_object()?
            .get("matches")?
            .as_number()
            .unwrap()
            .as_i64()
    })();
    ret.unwrap_or(0) as u64
}

pub fn show_rg(regex: &str) {
    Exec::cmd("rg")
        .arg("-i")
        .arg("--multiline")
        .arg(regex)
        .popen()
        .unwrap()
        .wait()
        .unwrap();
}

pub fn show_todos() {
    println!("{} todos found.", todos_amount());
    show_rg(TODO_REGEX);
}

fn main() {
    let args = args::get_args();

    match args.subcommand() {
        Some(("bootstrap", bs_args)) => {
            let base = bs_args.get_one::<String>("BASE").unwrap();
            let repo = bs_args.get_one::<String>("NEW").unwrap();
            bootstrap(base, repo);
        }
        Some(("new", new_args)) => {
            let branch: &String = new_args.get_one("BRANCH").unwrap();
            create_new_branch(branch);
        }
        Some(("remove", rm_args)) => {
            let branch: &String = rm_args.get_one("BRANCH").unwrap();
            delete_branch(branch);
        }
        Some(("branch", _)) => {
            Exec::cmd("git")
                .arg("branch")
                .popen()
                .unwrap()
                .wait()
                .unwrap();
        }
        Some(("push", push_args)) => {
            let force = push_args.get_flag("force");
            push(force);
        }
        Some(("commit", commit_args)) => {
            let all = commit_args.get_flag("all");
            let done = commit_args.get_flag("done");
            let msg: Option<&String> = commit_args.get_one("message");
            let interactive = commit_args.get_flag("interactive");

            if interactive || msg.is_none() {
                let gitmoji = select_gitmoji();
                let title = inquire::prompt_text("Commit Title").unwrap();
                let body = inquire::Text::new("Commit Description")
                    .with_page_size(5)
                    .prompt();
                let msg = CommitMessage::new(gitmoji, title, body.ok()).to_msg();
                commit(all, done, &msg);
            } else {
                commit(all, done, msg.unwrap());
            }
        }
        Some(("save", _)) => {
            commit(true, false, "WIP");
        }
        Some(("fetch", _)) => {
            fetch();
        }
        Some(("pull", _)) => {
            pull();
        }
        Some(("todo", _)) => {
            show_todos();
        }
        Some(("log", _)) => {
            Exec::cmd("serie").popen().unwrap().wait().unwrap();
        }
        Some(("status", _)) => {
            Exec::cmd("git")
                .arg("status")
                .popen()
                .unwrap()
                .wait()
                .unwrap();
        }
        Some(("add", args)) => {
            let file: &String = args.get_one("FILE").unwrap();
            git_add(file);
        }
        _ => {}
    }
}
